@@ -1,5 +1,6 @@
 /*
- * uconv version 0.4.12
+ * uconv version 0.5.3
+ * Jan  3, 2010 yoshidam  for ruby 1.9.1
  * Mar 12, 2003 yoshidam  for ruby 1.8.0
  * Sep  4, 2002 yoshidam  fixes memory leaks
  * Dec 10, 2001 yoshidam  supports the tainted status
@@ -13,11 +14,71 @@
  */
 
 #include "ruby.h"
-#include "rubyio.h"
+#ifdef HAVE_RUBY_IO_H
+#  include "ruby/io.h"
+#else
+#  include "rubyio.h"
+#endif
 #include <stdio.h>
+#include <string.h>
 #include "uconv.h"
 #include "ustring.h"
 
+#ifndef RSTRING_PTR
+#  define RSTRING_PTR(s) (RSTRING(s)->ptr)
+#  define RSTRING_LEN(s) (RSTRING(s)->len)
+#endif
+
+#ifdef HAVE_RUBY_ENCODING_H
+static rb_encoding* enc_u8;
+static rb_encoding* enc_u16le;
+static rb_encoding* enc_u16be;
+static rb_encoding* enc_u32le;
+static rb_encoding* enc_u32be;
+static rb_encoding* enc_euc;
+static rb_encoding* enc_sjis;
+static rb_encoding* enc_8bit;
+#  define ENC_U8(o) (rb_enc_associate(o, enc_u8))
+#  define ENC_U16LE(o) (rb_enc_associate(o, enc_u16le))
+#  define ENC_U16BE(o) (rb_enc_associate(o, enc_u16be))
+#  define ENC_U32LE(o) (rb_enc_associate(o, enc_u32le))
+#  define ENC_U32BE(o) (rb_enc_associate(o, enc_u32b))
+#  define ENC_EUC(o) (rb_enc_associate(o, enc_euc))
+#  define ENC_SJIS(o) (rb_enc_associate(o, enc_sjis))
+static inline VALUE
+enc_utf16swap(VALUE o, VALUE src) {
+  int src_enc = rb_enc_get_index(src);
+  if (src_enc == rb_enc_to_index(enc_u16le))
+    return rb_enc_associate(o, enc_u16be);
+  if (src_enc == rb_enc_to_index(enc_u16be))
+    return rb_enc_associate(o, enc_u16le);
+  else
+    return rb_enc_associate(o, enc_8bit);
+}
+static inline VALUE
+enc_utf32swap(VALUE o, VALUE src) {
+  int src_enc = rb_enc_get_index(src);
+  if (src_enc == rb_enc_to_index(enc_u32le))
+    return rb_enc_associate(o, enc_u32be);
+  if (src_enc == rb_enc_to_index(enc_u32be))
+    return rb_enc_associate(o, enc_u32le);
+  else
+    return rb_enc_associate(o, enc_8bit);
+}
+#  define ENC_U16SWAP(o, src) enc_utf16swap(o, src)
+#  define ENC_U32SWAP(o, src) enc_utf32swap(o, src)
+#else
+#  define ENC_U8(o) (o)
+#  define ENC_U16LE(o) (o)
+#  define ENC_U16BE(o) (o)
+#  define ENC_U32LE(o) (o)
+#  define ENC_U32LE(o) (o)
+#  define ENC_EUC(o) (o)
+#  define ENC_SJIS(o) (o)
+#  define ENC_U16SWAP(o, src) (o)
+#  define ENC_U32SWAP(o, src) (o)
+#endif
+
 /*#define UTF32*/
 
 #define REPLACEMENT_CHAR (0xFFFD)
@@ -29,6 +90,18 @@ static int shortest_flag = 1;
 static int replace_invalid = 0;
 static VALUE eUconvError;
 
+static ID id_unicode_eucjp_hook;
+static ID id_unicode_sjis_hook;
+static ID id_eucjp_hook;
+static ID id_sjis_hook;
+
+static ID id_unknown_euc_handler;
+static ID id_unknown_sjis_handler;
+static ID id_unknown_unicode_eucjp_handler;
+static ID id_unknown_unicode_sjis_handler;
+static ID id_unknown_unicode_handler;
+
+
 /* Convert UTF-8 to UTF-16-LE (byte order: 21) */
 static int
 _u8tou16(unsigned char* in, UString* out)
@@ -37,7 +110,7 @@ _u8tou16(unsigned char* in, UString* out)
   size_t inlen;
 
   UStr_alloc(out);
-  inlen = strlen(in);
+  inlen = strlen((char*)in);
   while (inlen > 0) {
     unsigned char c = in[0];
     if ((c & 0x80) == 0) { /* 0b0nnnnnnn (7bit) */
@@ -94,7 +167,7 @@ _u8tou16(unsigned char* in, UString* out)
 	}
 	else {
 	  UStr_free(out);
-	  rb_raise(eUconvError, "none-UTF-16 char detected (%04x)", u);
+	  rb_raise(eUconvError, "non-UTF-16 char detected (%04x)", u);
 	}
       }
       UStr_addChar2(out, u & 0xff, u >> 8);
@@ -136,10 +209,9 @@ _u8tou16(unsigned char* in, UString* out)
 	}
 	else {
 	  UStr_free(out);
-	  rb_raise(eUconvError, "none-UTF-16 char detected (%04x)", u);
+	  rb_raise(eUconvError, "non-UTF-16 char detected (%04x)", u);
 	}
       }
-      //      UStr_addWCharToU16LE(out, u);
       in += 4;
       inlen -= 4;
     }
@@ -167,7 +239,7 @@ _u8tou4(unsigned char* in, UString* out)
   size_t inlen;
 
   UStr_alloc(out);
-  inlen = strlen(in);
+  inlen = strlen((char*)in);
   while (inlen > 0) {
     unsigned char c = in[0];
     if ((c & 0x80) == 0) { /* 0b0nnnnnnn (7bit) */
@@ -224,7 +296,7 @@ _u8tou4(unsigned char* in, UString* out)
 	}
 	else {
 	  UStr_free(out);
-	  rb_raise(eUconvError, "none-UTF-16 char detected (%04x)", u);
+	  rb_raise(eUconvError, "non-UTF-16 char detected (%04x)", u);
 	}
       }
       UStr_addChar4(out, u & 0xff, u >> 8, 0, 0);
@@ -266,7 +338,7 @@ _u8tou4(unsigned char* in, UString* out)
 	}
 	else {
 	  UStr_free(out);
-	  rb_raise(eUconvError, "none-UTF-16 char detected (%04x)", u);
+	  rb_raise(eUconvError, "non-UTF-16 char detected (%04x)", u);
 	}
       }
 #endif /* UTF32 */
@@ -521,7 +593,7 @@ _u4tou16(unsigned char* in, int len, UString* out)
       }
       else {
 	UStr_free(out);
-	rb_raise(eUconvError, "none-UTF-16 char detected (%04x)", c);
+	rb_raise(eUconvError, "non-UTF-16 char detected (%04x)", c);
       }
     }
     else if (c < 0x10000) {     /* 0x0000-0xFFFF */
@@ -613,13 +685,8 @@ exception_handler(VALUE args, VALUE errinfo) {
 static VALUE
 call_euc_handler(const unsigned char* seq)
 {
-  ID mid = rb_intern("unknown_euc_handler");
-  VALUE ret;
-
-  if (!rb_method_boundp(CLASS_OF(mUconv), mid, 0))
-    return INT2FIX('?');
-
-  ret = rb_funcall((VALUE)mUconv, mid, 1, rb_str_new2((char*)seq));
+  VALUE ret = rb_funcall((VALUE)mUconv,
+                         id_unknown_euc_handler, 1, rb_str_new2((char*)seq));
   Check_Type(ret, T_FIXNUM);
   return ret;
 }
@@ -630,19 +697,47 @@ unknown_euc_handler(const unsigned char* seq)
   return rb_rescue((VALUE (*)(void*))call_euc_handler, (int)seq,
 		   exception_handler, Qnil);
 }
+
+static unknown_euc*
+check_euc_handler()
+{
+  if (!rb_method_boundp(CLASS_OF(mUconv), id_unknown_euc_handler, 0))
+    return NULL;
+  return unknown_euc_handler;
+}
+
+static VALUE
+call_e2u_hook(const unsigned char* seq)
+{
+  VALUE ret = rb_funcall((VALUE)mUconv,
+                         id_eucjp_hook, 1, rb_str_new2((char*)seq));
+  if (ret != Qnil)
+    Check_Type(ret, T_FIXNUM);
+  return ret;
+}
+
+static VALUE
+e2u_hook(const unsigned char* seq)
+{
+  return rb_rescue((VALUE (*)(void*))call_e2u_hook, (int)seq,
+		   exception_handler, Qnil);
+}
+
+static unknown_sjis*
+check_e2u_hook()
+{
+  if (!rb_method_boundp(CLASS_OF(mUconv), id_eucjp_hook, 0))
+    return NULL;
+  return e2u_hook;
+}
 #endif /* USE_EUC */
 
 #ifdef USE_SJIS
 static VALUE
 call_sjis_handler(const unsigned char* seq)
 {
-  ID mid = rb_intern("unknown_sjis_handler");
-  VALUE ret;
-
-  if (!rb_method_boundp(CLASS_OF(mUconv), mid, 0))
-    return INT2FIX('?');
-
-  ret = rb_funcall((VALUE)mUconv, mid, 1, rb_str_new2((char*)seq));
+  VALUE ret = rb_funcall((VALUE)mUconv,
+                         id_unknown_sjis_handler, 1, rb_str_new2((char*)seq));
   Check_Type(ret, T_FIXNUM);
   return ret;
 }
@@ -653,29 +748,162 @@ unknown_sjis_handler(const unsigned char* seq)
   return rb_rescue((VALUE (*)(void*))call_sjis_handler, (int)seq,
 		   exception_handler, Qnil);
 }
+
+static unknown_sjis*
+check_sjis_handler()
+{
+  if (!rb_method_boundp(CLASS_OF(mUconv), id_unknown_sjis_handler, 0))
+    return NULL;
+  return unknown_sjis_handler;
+}
+
+static VALUE
+call_s2u_hook(const unsigned char* seq)
+{
+  VALUE ret = rb_funcall((VALUE)mUconv,
+                         id_sjis_hook, 1, rb_str_new2((char*)seq));
+  if (ret != Qnil)
+    Check_Type(ret, T_FIXNUM);
+  return ret;
+}
+
+static VALUE
+s2u_hook(const unsigned char* seq)
+{
+  return rb_rescue((VALUE (*)(void*))call_s2u_hook, (int)seq,
+		   exception_handler, Qnil);
+}
+
+static unknown_sjis*
+check_s2u_hook()
+{
+  if (!rb_method_boundp(CLASS_OF(mUconv), id_sjis_hook, 0))
+    return NULL;
+  return s2u_hook;
+}
 #endif /* USE_SJIS */
 
 static VALUE
-call_unicode_handler(unsigned short code)
+call_unicode_handler(unsigned long code)
 {
-  ID mid = rb_intern("unknown_unicode_handler");
-  VALUE ret;
-
-  if (!rb_method_boundp(CLASS_OF(mUconv), mid, 0)) {
-    return rb_str_new2("?");
-  }
-  ret = rb_funcall((VALUE)mUconv, mid, 1, INT2FIX(code));
+  VALUE ret = rb_funcall((VALUE)mUconv,
+                         id_unknown_unicode_handler, 1, INT2FIX(code));
   Check_Type(ret, T_STRING);
   return ret;
 }
 
 static VALUE
-unknown_unicode_handler(unsigned short code)
+unknown_unicode_handler(unsigned long code)
 {
   return rb_rescue((VALUE (*)(void*))call_unicode_handler, code,
 		   exception_handler, Qnil);
 }
 
+static VALUE
+call_u2e_handler(unsigned long code)
+{
+  VALUE ret = rb_funcall((VALUE)mUconv,
+                         id_unknown_unicode_eucjp_handler, 1, INT2FIX(code));
+  Check_Type(ret, T_STRING);
+  return ret;
+}
+
+static VALUE
+unknown_u2e_handler(unsigned long code)
+{
+  return rb_rescue((VALUE (*)(void*))call_u2e_handler, code,
+		   exception_handler, Qnil);
+}
+
+static VALUE
+call_u2s_handler(unsigned long code)
+{
+  VALUE ret = rb_funcall((VALUE)mUconv,
+                         id_unknown_unicode_sjis_handler, 1, INT2FIX(code));
+  Check_Type(ret, T_STRING);
+  return ret;
+}
+
+static VALUE
+unknown_u2s_handler(unsigned long code)
+{
+  return rb_rescue((VALUE (*)(void*))call_u2s_handler, code,
+		   exception_handler, Qnil);
+}
+
+static unknown_unicode*
+check_unknown_u2e_handler()
+{
+  if (rb_method_boundp(CLASS_OF(mUconv), id_unknown_unicode_eucjp_handler, 0))
+    return unknown_u2e_handler;
+  else if (rb_method_boundp(CLASS_OF(mUconv), id_unknown_unicode_handler, 0))
+    return unknown_unicode_handler;
+  return NULL;
+}
+
+static unknown_unicode*
+check_unknown_u2s_handler()
+{
+  if (rb_method_boundp(CLASS_OF(mUconv), id_unknown_unicode_sjis_handler, 0))
+    return unknown_u2s_handler;
+  else if (rb_method_boundp(CLASS_OF(mUconv), id_unknown_unicode_handler, 0))
+    return unknown_unicode_handler;
+  return NULL;
+}
+
+static VALUE
+call_u2e_hook(unsigned long code)
+{
+  VALUE ret;
+
+  ret = rb_funcall((VALUE)mUconv, id_unicode_eucjp_hook, 1, INT2FIX(code));
+  if (ret != Qnil)
+    Check_Type(ret, T_STRING);
+  return ret;
+}
+
+static VALUE
+u2e_hook(unsigned long code)
+{
+  return rb_rescue((VALUE (*)(void*))call_u2e_hook, code,
+		   exception_handler, Qnil);
+}
+
+static unknown_unicode*
+check_u2e_hook()
+{
+  if (!rb_method_boundp(CLASS_OF(mUconv), id_unicode_eucjp_hook, 0))
+    return NULL;
+  return u2e_hook;
+}
+
+static VALUE
+call_u2s_hook(unsigned long code)
+{
+  VALUE ret;
+
+  ret = rb_funcall((VALUE)mUconv, id_unicode_sjis_hook, 1, INT2FIX(code));
+  if (ret != Qnil)
+    Check_Type(ret, T_STRING);
+  return ret;
+}
+
+static VALUE
+u2s_hook(unsigned long code)
+{
+  return rb_rescue((VALUE (*)(void*))call_u2s_hook, code,
+		   exception_handler, Qnil);
+}
+
+static unknown_unicode*
+check_u2s_hook()
+{
+  if (!rb_method_boundp(CLASS_OF(mUconv), id_unicode_sjis_hook, 0))
+    return NULL;
+  return u2s_hook;
+}
+
+
 #ifdef USE_EUC
 static VALUE
 uconv_u2toeuc(VALUE obj, VALUE wstr)
@@ -687,10 +915,10 @@ uconv_u2toeuc(VALUE obj, VALUE wstr)
 
   Check_Type(wstr, T_STRING);
 
-  u = (unsigned char*)(RSTRING(wstr)->ptr);
-  len = RSTRING(wstr)->len;
+  u = (unsigned char*)(RSTRING_PTR(wstr));
+  len = RSTRING_LEN(wstr);
   if (u) {
-    u2e_conv2(u, len, &e, unknown_unicode_handler);
+    u2e_conv2(u, len, &e, check_unknown_u2e_handler(), check_u2e_hook());
     ret = rb_str_new((char*)e.str, e.len);
     UStr_free(&e);
   }
@@ -712,9 +940,9 @@ uconv_euctou2(VALUE obj, VALUE estr)
 
   Check_Type(estr, T_STRING);
 
-  e = (unsigned char*)(RSTRING(estr)->ptr);
+  e = (unsigned char*)(RSTRING_PTR(estr));
   if (e) {
-    len = e2u_conv2(e, &u, unknown_euc_handler);
+    len = e2u_conv2(e, &u, check_euc_handler(), check_e2u_hook());
 
     ret = rb_str_new((char*)u.str, u.len);
     UStr_free(&u);
@@ -722,7 +950,7 @@ uconv_euctou2(VALUE obj, VALUE estr)
   else {
     ret = rb_str_new(NULL, 0);
   }
-  OBJ_INFECT(ret, estr);
+  OBJ_INFECT(ENC_U16LE(ret), estr);
 
   return ret;
 }
@@ -739,18 +967,18 @@ uconv_u2tosjis(VALUE obj, VALUE wstr)
 
   Check_Type(wstr, T_STRING);
 
-  u = (unsigned char*)(RSTRING(wstr)->ptr);
+  u = (unsigned char*)(RSTRING_PTR(wstr));
   if (u) {
-    len = RSTRING(wstr)->len;
+    len = RSTRING_LEN(wstr);
 
-    u2s_conv2(u, len, &s, unknown_unicode_handler);
+    u2s_conv2(u, len, &s, check_unknown_u2s_handler(), check_u2s_hook());
     ret = rb_str_new((char*)s.str, s.len);
     UStr_free(&s);
   }
   else {
     ret = rb_str_new(NULL, 0);
   }
-  OBJ_INFECT(ret, wstr);
+  OBJ_INFECT(ENC_SJIS(ret), wstr);
 
   return ret;
 }
@@ -765,9 +993,9 @@ uconv_sjistou2(VALUE obj, VALUE sstr)
 
   Check_Type(sstr, T_STRING);
 
-  s = (unsigned char*)(RSTRING(sstr)->ptr);
+  s = (unsigned char*)(RSTRING_PTR(sstr));
   if (s) {
-    len = s2u_conv2(s, &u, unknown_sjis_handler);
+    len = s2u_conv2(s, &u, check_sjis_handler(), check_s2u_hook());
 
     ret = rb_str_new((char*)u.str, u.len);
     UStr_free(&u);
@@ -775,7 +1003,7 @@ uconv_sjistou2(VALUE obj, VALUE sstr)
   else {
     ret = rb_str_new(NULL, 0);
   }
-  OBJ_INFECT(ret, sstr);
+  OBJ_INFECT(ENC_U16LE(ret), sstr);
 
   return ret;
 }
@@ -792,11 +1020,11 @@ uconv_u2swap(VALUE obj, VALUE wstr)
 
   Check_Type(wstr, T_STRING);
 
-  u = (unsigned char*)(RSTRING(wstr)->ptr);
-  len = RSTRING(wstr)->len;
+  u = (unsigned char*)(RSTRING_PTR(wstr));
+  len = RSTRING_LEN(wstr);
   if (!u || len < 2) return Qnil;
   ret = rb_str_new(NULL, len);
-  out = (unsigned char*)(RSTRING(ret)->ptr);
+  out = (unsigned char*)(RSTRING_PTR(ret));
 
   for (i = 0; i < len; i+=2) {
     out[i] = u[i+1];
@@ -804,7 +1032,7 @@ uconv_u2swap(VALUE obj, VALUE wstr)
   }
   OBJ_INFECT(ret, wstr);
 
-  return ret;
+  return ENC_U16SWAP(ret, wstr);
 }
 
 static VALUE
@@ -817,8 +1045,8 @@ uconv_u2swap_b(VALUE obj, VALUE wstr)
   Check_Type(wstr, T_STRING);
 
   rb_str_modify(wstr);
-  u = (unsigned char*)(RSTRING(wstr)->ptr);
-  len = RSTRING(wstr)->len;
+  u = (unsigned char*)(RSTRING_PTR(wstr));
+  len = RSTRING_LEN(wstr);
   if (!u || len < 2) return Qnil;
 
   for (i = 0; i < len; i+=2) {
@@ -826,7 +1054,7 @@ uconv_u2swap_b(VALUE obj, VALUE wstr)
     u[i+1] = u[i];
     u[i] = tmp;
   }
-  return wstr;
+  return ENC_U16SWAP(wstr, wstr);
 }
 
 static VALUE
@@ -840,11 +1068,11 @@ uconv_u4swap(VALUE obj, VALUE wstr)
 
   Check_Type(wstr, T_STRING);
 
-  u = (unsigned char*)(RSTRING(wstr)->ptr);
-  len = RSTRING(wstr)->len;
+  u = (unsigned char*)(RSTRING_PTR(wstr));
+  len = RSTRING_LEN(wstr);
   if (!u || len < 4) return Qnil;
   ret = rb_str_new(NULL, len);
-  out = (unsigned char*)(RSTRING(ret)->ptr);
+  out = (unsigned char*)(RSTRING_PTR(ret));
 
   for (i = 0; i < len; i+=4) {
     out[i]   = u[i+3];
@@ -854,7 +1082,7 @@ uconv_u4swap(VALUE obj, VALUE wstr)
   }
   OBJ_INFECT(ret, wstr);
 
-  return ret;
+  return ENC_U32SWAP(ret, wstr);
 }
 
 static VALUE
@@ -867,8 +1095,8 @@ uconv_u4swap_b(VALUE obj, VALUE wstr)
   Check_Type(wstr, T_STRING);
 
   rb_str_modify(wstr);
-  u = (unsigned char*)(RSTRING(wstr)->ptr);
-  len = RSTRING(wstr)->len;
+  u = (unsigned char*)(RSTRING_PTR(wstr));
+  len = RSTRING_LEN(wstr);
   if (len < 4) return Qnil;
 
   for (i = 0; i < len; i+=4) {
@@ -879,7 +1107,7 @@ uconv_u4swap_b(VALUE obj, VALUE wstr)
     u[i+2] = tmp2;
     u[i+3] = tmp1;
   }
-  return wstr;
+  return ENC_U32SWAP(wstr, wstr);
 }
 
 
@@ -891,7 +1119,7 @@ uconv_u8tou16(VALUE obj, VALUE ustr)
   VALUE ret;
 
   Check_Type(ustr, T_STRING);
-  in = (unsigned char*)(RSTRING(ustr)->ptr);
+  in = (unsigned char*)(RSTRING_PTR(ustr));
   if (in) {
     _u8tou16(in, &out);
     ret = rb_str_new((char*)out.str, out.len);
@@ -900,7 +1128,7 @@ uconv_u8tou16(VALUE obj, VALUE ustr)
   else {
     ret = rb_str_new(NULL, 0);
   }
-  OBJ_INFECT(ret, ustr);
+  OBJ_INFECT(ENC_U16LE(ret), ustr);
 
   return ret;
 }
@@ -914,9 +1142,9 @@ uconv_u16tou8(VALUE obj, VALUE wstr)
   VALUE ret;
 
   Check_Type(wstr, T_STRING);
-  in = (unsigned char*)(RSTRING(wstr)->ptr);
+  in = (unsigned char*)(RSTRING_PTR(wstr));
   if (in) {
-    len = RSTRING(wstr)->len;
+    len = RSTRING_LEN(wstr);
     _u16tou8(in, len, &out, eliminate_zwnbsp_flag);
     ret = rb_str_new((char*)out.str, out.len);
     UStr_free(&out);
@@ -924,7 +1152,7 @@ uconv_u16tou8(VALUE obj, VALUE wstr)
   else {
     ret = rb_str_new(NULL, 0);
   }
-  OBJ_INFECT(ret, wstr);
+  OBJ_INFECT(ENC_U8(ret), wstr);
 
   return ret;
 }
@@ -938,7 +1166,7 @@ uconv_u8tou4(VALUE obj, VALUE ustr)
   VALUE ret;
 
   Check_Type(ustr, T_STRING);
-  in = (unsigned char*)(RSTRING(ustr)->ptr);
+  in = (unsigned char*)(RSTRING_PTR(ustr));
   if (in) {
     _u8tou4(in, &out);
     ret = rb_str_new((char*)out.str, out.len);
@@ -947,7 +1175,7 @@ uconv_u8tou4(VALUE obj, VALUE ustr)
   else {
     ret = rb_str_new(NULL, 0);
   }
-  OBJ_INFECT(ret, ustr);
+  OBJ_INFECT(ENC_U32LE(ret), ustr);
 
   return ret;
 }
@@ -961,9 +1189,9 @@ uconv_u4tou8(VALUE obj, VALUE wstr)
   VALUE ret;
 
   Check_Type(wstr, T_STRING);
-  in = (unsigned char*)(RSTRING(wstr)->ptr);
+  in = (unsigned char*)(RSTRING_PTR(wstr));
   if (in) {
-    len = RSTRING(wstr)->len;
+    len = RSTRING_LEN(wstr);
     _u4tou8(in, len, &out, eliminate_zwnbsp_flag);
     ret = rb_str_new((char*)out.str, out.len);
     UStr_free(&out);
@@ -971,7 +1199,7 @@ uconv_u4tou8(VALUE obj, VALUE wstr)
   else {
     ret = rb_str_new(NULL, 0);
   }
-  OBJ_INFECT(ret, wstr);
+  OBJ_INFECT(ENC_U8(ret), wstr);
 
   return ret;
 }
@@ -986,9 +1214,9 @@ uconv_u16tou4(VALUE obj, VALUE wstr)
   VALUE ret;
 
   Check_Type(wstr, T_STRING);
-  in = (unsigned char*)(RSTRING(wstr)->ptr);
+  in = (unsigned char*)(RSTRING_PTR(wstr));
   if (in) {
-    len = RSTRING(wstr)->len;
+    len = RSTRING_LEN(wstr);
     _u16tou4(in, len, &out);
     ret = rb_str_new((char*)out.str, out.len);
     UStr_free(&out);
@@ -996,7 +1224,7 @@ uconv_u16tou4(VALUE obj, VALUE wstr)
   else {
     ret = rb_str_new(NULL, 0);
   }
-  OBJ_INFECT(ret, wstr);
+  OBJ_INFECT(ENC_U32LE(ret), wstr);
 
   return ret;
 }
@@ -1010,9 +1238,9 @@ uconv_u4tou16(VALUE obj, VALUE wstr)
   VALUE ret;
 
   Check_Type(wstr, T_STRING);
-  in = (unsigned char*)(RSTRING(wstr)->ptr);
+  in = (unsigned char*)(RSTRING_PTR(wstr));
   if (in) {
-    len = RSTRING(wstr)->len;
+    len = RSTRING_LEN(wstr);
     _u4tou16(in, len, &out);
     ret = rb_str_new((char*)out.str, out.len);
     UStr_free(&out);
@@ -1020,7 +1248,7 @@ uconv_u4tou16(VALUE obj, VALUE wstr)
   else {
     ret = rb_str_new(NULL, 0);
   }
-  OBJ_INFECT(ret, wstr);
+  OBJ_INFECT(ENC_U16LE(ret), wstr);
 
   return ret;
 }
@@ -1036,20 +1264,20 @@ uconv_u8toeuc(VALUE obj, VALUE ustr)
   volatile VALUE vout;
 
   Check_Type(ustr, T_STRING);
-  in = (unsigned char*)(RSTRING(ustr)->ptr);
+  in = (unsigned char*)(RSTRING_PTR(ustr));
   if (in) {
     _u8tou16(in, &out);
     vout = rb_str_new((char*)out.str, out.len);
     UStr_free(&out);
-    u2e_conv2(RSTRING(vout)->ptr, RSTRING(vout)->len,
-	      &e, unknown_unicode_handler);
+    u2e_conv2((unsigned char*)(RSTRING_PTR(vout)), RSTRING_LEN(vout),
+	      &e, check_unknown_u2e_handler(), check_u2e_hook());
     ret = rb_str_new((char*)e.str, e.len);
     UStr_free(&e);
   }
   else {
     ret = rb_str_new(NULL, 0);
   }
-  OBJ_INFECT(ret, ustr);
+  OBJ_INFECT(ENC_EUC(ret), ustr);
 
   return ret;
 }
@@ -1064,19 +1292,19 @@ uconv_euctou8(VALUE obj, VALUE estr)
   VALUE vin;
 
   Check_Type(estr, T_STRING);
-  e = (unsigned char*)(RSTRING(estr)->ptr);
+  e = (unsigned char*)(RSTRING_PTR(estr));
   if (e) {
-    e2u_conv2(e, &in, unknown_euc_handler);
+    e2u_conv2(e, &in, check_euc_handler(), check_e2u_hook());
     vin = rb_str_new((char*)in.str, in.len);
     UStr_free(&in);
-    _u16tou8(RSTRING(vin)->ptr, RSTRING(vin)->len, &out, 1);
+    _u16tou8((unsigned char*)(RSTRING_PTR(vin)), RSTRING_LEN(vin), &out, 1);
     ret = rb_str_new((char*)out.str, out.len);
     UStr_free(&out);
   }
   else {
     ret = rb_str_new(NULL, 0);
   }
-  OBJ_INFECT(ret, estr);
+  OBJ_INFECT(ENC_U8(ret), estr);
 
   return ret;
 }
@@ -1093,20 +1321,20 @@ uconv_u8tosjis(VALUE obj, VALUE ustr)
   volatile VALUE vout;
 
   Check_Type(ustr, T_STRING);
-  in = (unsigned char*)(RSTRING(ustr)->ptr);
+  in = (unsigned char*)(RSTRING_PTR(ustr));
   if (in) {
     _u8tou16(in, &out);
     vout = rb_str_new((char*)out.str, out.len);
     UStr_free(&out);
-    u2s_conv2(RSTRING(vout)->ptr, RSTRING(vout)->len,
-	      &s, unknown_unicode_handler);
+    u2s_conv2((unsigned char*)(RSTRING_PTR(vout)), RSTRING_LEN(vout),
+	      &s, check_unknown_u2s_handler(), check_u2s_hook());
     ret = rb_str_new((char*)s.str, s.len);
     UStr_free(&s);
   }
   else {
     ret = rb_str_new(NULL, 0);
   }
-  OBJ_INFECT(ret, ustr);
+  OBJ_INFECT(ENC_SJIS(ret), ustr);
 
   return ret;
 }
@@ -1121,19 +1349,19 @@ uconv_sjistou8(VALUE obj, VALUE sstr)
   VALUE vin;
 
   Check_Type(sstr, T_STRING);
-  s = (unsigned char*)(RSTRING(sstr)->ptr);
+  s = (unsigned char*)(RSTRING_PTR(sstr));
   if (s) {
-    s2u_conv2(s, &in, unknown_sjis_handler);
+    s2u_conv2(s, &in, check_sjis_handler(), check_s2u_hook());
     vin = rb_str_new((char*)in.str, in.len);
     UStr_free(&in);
-    _u16tou8(RSTRING(vin)->ptr, RSTRING(vin)->len, &out, 1);
+    _u16tou8((unsigned char*)(RSTRING_PTR(vin)), RSTRING_LEN(vin), &out, 1);
     ret = rb_str_new((char*)out.str, out.len);
     UStr_free(&out);
   }
   else {
     ret = rb_str_new(NULL, 0);
   }
-  OBJ_INFECT(ret, sstr);
+  OBJ_INFECT(ENC_U8(ret), sstr);
 
   return ret;
 }
@@ -1205,6 +1433,17 @@ set_replace_invalid(VALUE obj, VALUE uchar)
 void
 Init_uconv()
 {
+#ifdef HAVE_RUBY_ENCODING_H
+  enc_u8 = rb_utf8_encoding();
+  enc_u16le = rb_enc_find("UTF-16LE");
+  enc_u16be = rb_enc_find("UTF-16BE");
+  enc_u32le = rb_enc_find("UTF-32LE");
+  enc_u32be = rb_enc_find("UTF-32BE");
+  enc_euc = rb_enc_find("EUC-JP");
+  enc_sjis = rb_enc_find("Windows-31J");
+  enc_8bit = rb_ascii8bit_encoding();
+#endif
+
   if (rb_const_defined(rb_cObject, rb_intern("Uconv")) == Qtrue)
     mUconv = rb_const_get(rb_cObject, rb_intern("Uconv"));
   else
@@ -1212,6 +1451,16 @@ Init_uconv()
   eUconvError = rb_define_class_under(mUconv,
 				      "Error", rb_eStandardError);
 
+  id_unicode_eucjp_hook = rb_intern("unicode_euc_hook");
+  id_unicode_sjis_hook = rb_intern("unicode_sjis_hook");
+  id_eucjp_hook = rb_intern("euc_hook");
+  id_sjis_hook = rb_intern("sjis_hook");
+  id_unknown_euc_handler = rb_intern("unknown_euc_handler");
+  id_unknown_sjis_handler = rb_intern("unknown_sjis_handler");
+  id_unknown_unicode_handler = rb_intern("unknown_unicode_handler");
+  id_unknown_unicode_eucjp_handler = rb_intern("unknown_unicode_euc_handler");
+  id_unknown_unicode_sjis_handler = rb_intern("unknown_unicode_sjis_handler");
+
 #ifdef USE_EUC
   rb_define_module_function(mUconv, "u16toeuc", uconv_u2toeuc, 1);
   rb_define_module_function(mUconv, "euctou16", uconv_euctou2, 1);
